/*
 * Copyright (c) 2005 TopCoder, Inc., All Rights Reserved
 */
package com.topcoder.network.synchronization.defaultimpl;

import java.util.Hashtable;
import java.util.Map;

import com.topcoder.network.synchronization.protocol.PeerGroupService;


/**
 * This class extending PeerGroupService provides synchronized peer group service.
 * Compare to the common peer group service, this service ensure that each peer in the peer group processes
 * the message in the same order. For example, peer A sends a message MA and peer B sends a message MB, each
 * peer in the group will process them in MA->MB order or each peer will process them in MB->MA order.
 * 
 * So that, this class adds some elements to solve it. The most important two are dispatcher and synchronization controller.
 * Dispatcher dispatchs messages from receiver to different message containers, and controller communicates with other peers
 * to synchronize the message process.
 * 
 * To keep this peer group service alive, the KeepAliveEngine is used. It sends keep alive messages every an interval time, 
 * and judges the service to die if less than a predefined peers had replied the keep alive message.
 * 
 * @author TCSDESIGNER
 * @version 1.0
 */
public class SynchronizedPeerGroupService extends PeerGroupService {
    
    /**
     * Represents the state of this service.
     */
    protected ServiceState state = null;    
       
    /**
     * Represents the message queue to store all the messages sended by the user application.
     */
    protected MessageQueue actionMessageQueue = null;
    
    /**
     * Represents all the messages(APPLICATION_MESSAGE & COMMUNICATE_MESSAGE) received in this service,
     * but if the message is already dealed, remove the message from it. 
     */
    protected Map allMessages = new Hashtable();
    
    /**
     * Represents the COMMUNICATE_MESSAGEs in the current turn.
     */
    protected Map communicateMessages = new Hashtable();   
    
    /**
     * Represents the KEEPALIVE_MESSAGEs in the current turn.
     */
    protected Map keepAliveMessages = new Hashtable();
    
    /**
     * The engine is used to ensure this peer is alive in the current peer group.
     */
    protected KeepAliveEngine keepAlive = null;    
    
    /**
     * Constructor.
     *
     */
    public SynchronizedPeerGroupService() {
    
    }
    
    /**
     * 1. start message sender.
     * 2. add ReplyJoinMessageListener to the receiver.
     * 3. start the receiver.
     * 4. send the JOINGROUP_MESSAGE
     * 5. receive the reply from the other peers in the group
     * 6. if successful- remove the ReplyJoinMessageListener.
     * 6.1 	add ForwardMessageListener to the receiver
     * 6.2 	add RequestMessageListener to the receiver
     * 6.3  add CommunicationMessageListener
     * 6.4  add KeepAliveMessageLister.
     * 6.5  add ApplicationMessageLister.
     * 6.6  go into the synchronization controller loop.
     * 6.6.1 if the actionMessageQueue is not empty, send a COMMUNICATE_MESSAGE to vote
     *       a message in actionMessageQueue to process
     * 6.6.2 check the votes in the communicateMessages map
     * 6.6.3 if some votes are missed, send REQUEST_MESSAGE to request that vote from other peers.
     * 6.6.4 if some votes are still missed - 
     * 6.6.4.1 if the number of votes are equal or larger than the (number of peers in group) * keepAlivePersent,
     *         send a ERROR_MESSAGE to tell the missed peers that they are unsynchronized and should be stopped.
     * 		   And remove that peers from the peer group.
     * 6.6.4.2 otherwise, stop the service immediately
     * 6.6.5 if all votes are received, choose the message got most votes to process (remove it from the 
     *       actionMessagesQueue and pass it to the first ApplicationMessageLister.)
     * 6.6.6 goto 6.6.1
     *   
     */
    public void run() {
    
    }
    
    /**
     * 1. send LEAVEGROUP_MESSAGE to leave the current peer group
     * 2. stop sender, receiver and the synchronization cotroller loop.
     */
    public void stop() {
    
    }
    
    
}
